import sys

from encoders.basic import *
from formula import *
from structures import *

from errors import *
import common


class ClausalHelper(object):

    def __init__(self, problem, environment):
        self.environment = environment
        self.problem = problem
        self.psiClauses = []
        self.gammaClauses = []

        self.ybounds = {}
        self.ybases = {}

        self.yvars = []
        self.xvars = []

        self.atoms = []
        self.gamma_atoms = []

        self._prepare()
        return

    def _prepare(self):
        self.psiClauses = []
        self.gammaClauses = []

        st = self.environment.symbolTable

        for tp in self.problem.timePoints.values():
            if tp.controllable:
                x = st.getOrCreateSymbol(tp.name, Types.REAL)
                self.xvars.append(x)

        eNodes = {}
        for e in self.problem.constraints:
            if e.contingent:
                y = st.createNewFreshSymbol(Types.REAL, "_Y_%d")
                self.yvars.append(y)
                disj = []
                for a in e.atoms:
                    s = a.source
                    d = a.dest
                    l = a.lowerBound
                    u = a.upperBound

                    assert not d.controllable
                    assert s.controllable

                    vs = st.getSymbolByName(s.name)
                    assert vs is not None

                    lst = self.ybounds.get(y, [])
                    lst.append((0, u-l))
                    self.ybounds[y] = lst

                    oth = self.ybases.get(y, s)
                    assert oth == s
                    self.ybases[y] = s

                    vd = Minus(Plus(vs, Constant(u)), y)

                    eNodes[d.name] = vd

                    if l == common.MINUS_INFINITY:
                        common.error("A contingent constraint has a negative"+
                                     " lower or upper bound")

                    if u == common.INFINITY:
                        common.error("The nature can lock the system since an " +
                                     "uncontrollable interval has an infinite bound")

                    disj.append(And([GE(y, Constant(0)), LE(y, Constant(u-l))]))

                    self.gamma_atoms.append(GE(y, Constant(0)))
                    self.gamma_atoms.append(LE(y, Constant(u-l)))


                    print(y.name + " --> " + s.name)

                self.gammaClauses.append(Or(disj))


        for e in self.problem.constraints:
            if not e.contingent:
                c = self._constraint_2_clause(e, eNodes)
                self.psiClauses.append(c)
        return


    def _constraint_2_clause(self, constraint, eNodes):
        st = self.environment.symbolTable
        disj = []
        for a in constraint.atoms:
            s = a.source
            d = a.dest
            l = a.lowerBound
            u = a.upperBound

            vs = eNodes[s.name] if s.name in eNodes else st.getSymbolByName(s.name)
            vd = eNodes[d.name] if d.name in eNodes else st.getSymbolByName(d.name)

            conj = []
            if l != common.MINUS_INFINITY:
                conj.append(GE(Minus(vd,vs), Constant(l)))
            if u != common.INFINITY:
                conj.append(LE(Minus(vd,vs), Constant(u)))

            assert len(conj) > 0

            for c in conj:
                self.atoms.append(c)

            disj.append(And(conj))
        return Or(disj)

    def getGamma(self):
        gc = self.getGammaClauses()
        return And(gc)

    def getPsi(self):
        fc = self.getPsiClauses()
        return And(fc)

    def getGammaClauses(self):
        return self.gammaClauses

    def getPsiClauses(self):
        return self.psiClauses

    def getYBound(self, y):
        lst = self.ybounds[y]
        assert len(lst) == 1
        return lst[0]

    def getYBase(self, y):
        return self.ybases[y]
    pass


class VidalHelper(object):

    def __init__(self):
        self.quantifiedYvars = None
        return

    def isY(self, var):
        if self.quantifiedYvars is not None:
            return var in self.quantifiedYvars
        return var.nodeType == FNode.DEFINE

    def isX(self, var):
        if self.quantifiedYvars is not None:
            return var not in self.quantifiedYvars
        return var.nodeType == FNode.VARIABLE


    def getVidalElimination(self, c):
        strict = True
        if c.nodeType == FNode.LT and c.left.nodeType == FNode.MINUS and c.right.nodeType == FNode.CONSTANT:
            strict = True
            c = GT(Minus(c.left.right, c.left.left), Constant(-1 * c.right.value))
        if c.nodeType == FNode.LE and c.left.nodeType == FNode.MINUS and c.right.nodeType == FNode.CONSTANT:
            strict = False
            c = GE(Minus(c.left.right, c.left.left), Constant(-1 * c.right.value))
        else:
            assert c.nodeType == FNode.GT or c.nodeType == FNode.GE
            strict = c.nodeType == FNode.GT

        left = c.left.left
        right = c.left.right
        a = c.right.value

        assert self.isY(left) or self.isX(left)
        assert self.isY(right) or self.isX(right)

        if self.isX(left) and self.isX(right):
            #type 1
            return c
        else:
            lbody = None
            rbody = None
            if self.isY(left) and self.isX(right):
                #type2
                key = left.name
                (l,u) = self.getBounds(key)
                lbody = Minus(self.getBase(key), right)
                rbody = Constant(a-l)
            elif self.isX(left) and self.isY(right):
                #type3
                key = right.name
                (l,u) = self.getBounds(key)
                lbody = Minus(left, self.getBase(key))
                rbody = Constant(a+u)
            elif self.isY(left) and self.isY(right):
                #type4
                key1 = left.name
                (l1,u1) = self.getBounds(key1)
                key2 = right.name
                (l2,u2) = self.getBounds(key2)
                lbody = Minus(self.getBase(key1), self.getBase(key2))
                rbody = Constant(a-l1+u2)

            else:
                assert False
            if strict:
                return GT(lbody, rbody)
            else:
                return GE(lbody, rbody)

        assert False
        return
