from encoders.utils import *
from encoders.helpers import *
from extractors.strategy import *

from extractors.solvers.msat import *
from checkers import *

import common
import random

import sympy

def sympy2fraction(n):
    r = sympy.Rational(str(n))
    return Fraction(r.p, r.q)

def rank(M):
    return len(M.rref()[1])


class SkinStrategyExtractor(object):

    def __init__(self, environment):
        self.environment = environment
        return


    def getCoefficient(self, poly, varname):
        if poly.nodeType == FNode.SYMBOL:
            if poly.name == varname:
                return Fraction(1, 1)
            return Fraction(0, 1)

        if poly.nodeType == FNode.CONSTANT:
            return poly.value

        elif poly.nodeType == FNode.TIMES:
            return self.getCoefficient(poly.left, varname) * \
                self.getCoefficient(poly.right, varname)

        elif poly.nodeType in [FNode.PLUS, FNode.MINUS]:
            l = Fraction(0, 1)
            r = Fraction(0, 1)
            if poly.left.nodeType != FNode.CONSTANT:
                l = self.getCoefficient(poly.left, varname)

            if poly.right.nodeType != FNode.CONSTANT:
                r = self.getCoefficient(poly.right, varname)

            if poly.nodeType == FNode.PLUS:
                return l+r
            else:
                return l-r

        else:
            assert False


    def getBias(self, poly):
        if poly.nodeType == FNode.SYMBOL:
            return Fraction(0, 1)

        if poly.nodeType == FNode.CONSTANT:
            return poly.value

        elif poly.nodeType == FNode.TIMES:
            return self.getBias(poly.left) * \
                self.getBias(poly.right)

        elif poly.nodeType in [FNode.PLUS, FNode.MINUS]:
            l = self.getBias(poly.left)
            r = self.getBias(poly.right)

            if poly.nodeType == FNode.PLUS:
                return l+r
            else:
                return l-r

        else:
            assert False


    def getRow(self, equality, clausifier):
        assert equality.nodeType == FNode.EQUALS
        yvars = clausifier.yvars
        xvars = clausifier.xvars
        allvars = xvars + yvars

        left = equality.left
        right = equality.right
        if left.nodeType == FNode.CONSTANT:
            row = [self.getCoefficient(right, v.name) for v in allvars]
            row.append(left.value - self.getBias(right))
        else:
            assert right.nodeType == FNode.CONSTANT
            row = [self.getCoefficient(left, v.name) for v in allvars]
            row.append(right.value - self.getBias(left))

        return row


    def getLinearStrategyFromSystem(self, system, clausifier, xmodel):
        yvars = clausifier.yvars
        xvars = clausifier.xvars

        matrix = []
        if system.nodeType == FNode.EQUALS:
            row = self.getRow(system, clausifier)
            matrix.append([sympy.Rational(f.numerator, f.denominator) for f in row])
        else:
            assert system.nodeType == FNode.AND
            for c in system.conjunction:
                assert c.nodeType == FNode.EQUALS
                row = self.getRow(c, clausifier)
                matrix.append([sympy.Rational(f.numerator, f.denominator) for f in row])

        many = len(xvars)
        assert len(matrix) > 0
        if len(matrix) < many:
            for _ in xrange(many-len(matrix)):
                matrix.append([sympy.Rational(0, 1) for _ in xrange(len(yvars)+many+1)])

        m = sympy.Matrix(matrix)
        rref =  m.rref()[0]

        ker = len(xvars) + len(yvars) - rank(rref)

        idx = many
        cols = many + len(yvars) + 1

        values = {}
        for x in xvars:
            row = [Fraction(0,1) for _ in yvars] + [xmodel[x]]
            values[x] = row

        for _ in reversed(xrange(many)):
            idx -= 1
            j = idx
            while j<cols and rref[idx,j] == 0:
                j+= 1

            if j < many:
                row = None
                x = xvars[j]

                c = rref[idx,j]
                assert c == 1

                ter = rref[idx,cols-1]
                row = [-1 * sympy2fraction(rref[idx, w]) \
                            for w in xrange(many, cols-1)] + \
                            [sympy2fraction(ter)]

                for w in xrange(j+1, many):
                    mul = rref[idx,w]
                    if mul != 0:
                        tim = -1 * sympy2fraction(mul)
                        ox = xvars[w]
                        orl = values[ox]
                        nrl = [tim*f for f in orl]
                        row = [row[k] + nrl[k] for k in xrange(len(row))]

                assert row is not None
                values[x] = row

        s = LinearStrategy(yvars)
        for x in xvars:
            s.addPolicy(x, values[x])

        # Checker
        tmp = MathSAT5Solver(self.environment)
        for x in xvars + yvars:
            tmp.declareVariable(x)
        for y in yvars:
            tmp.addAssertion(Equals(y, Constant(xmodel[y])))
        for x in xvars:
            tmp.addAssertion(Equals(x, s.getPolicyAsFormula(x)))

        assert tmp.solve()
        tmp.destroy()

        return (s, ker)


    def getPsiWithStartegy(self, strategy, clausifier):
        xvars = clausifier.xvars
        formula = clausifier.getPsi()

        for x in xvars:
            formula = Definition(x.name, strategy.getPolicyAsFormula(x), formula)

        return (formula.simplify(True))


    def getInternals(self, formula):
        if formula.nodeType == FNode.AND:
            rewritten = []
            for son in formula.conjunction:
                rewritten.append(self.getInternals(son))
            return And(rewritten)

        elif formula.nodeType == FNode.OR:
            rewritten = []
            for son in formula.disjunction:
                rewritten.append(self.getInternals(son))
            return Or(rewritten)

        elif formula.nodeType == FNode.NOT:
            return Not(self.getInternals(formula.formula))

        elif formula.nodeType == FNode.GE:
            sl = formula.left
            sr = formula.right
            return GT(sl, sr)

        elif formula.nodeType == FNode.LE:
            sl = formula.left
            sr = formula.right
            return LT(sl, sr)

        elif formula.nodeType == FNode.FORALL:
            f = self.getInternals(formula.formula)
            return ForAll(formula.variables, f)

        elif formula.nodeType == FNode.EXISTS:
            f = self.getInternals(formula.formula)
            return Exists(formula.variables, f)

        elif formula.nodeType == FNode.DEFINITION:
            d = self.getInternals(formula.definition)
            f = self.getInternals(formula.formula)
            return Definition(formula.name, d, f)


        elif formula.nodeType == FNode.EQUALS:
            assert False

        elif formula.nodeType == FNode.TEXTFORMULA:
            assert False

        else:
            return formula


    def getEqualities(self, formula):
        if formula.nodeType == FNode.AND:
            res = []
            for son in formula.conjunction:
                res += self.getEqualities(son)
            return res

        elif formula.nodeType == FNode.OR:
            res = []
            for son in formula.disjunction:
                res += self.getEqualities(son)
            return res

        elif formula.nodeType == FNode.NOT:
            return self.getInternals(formula.formula)

        elif formula.nodeType == FNode.GE:
            sl = formula.left
            sr = formula.right
            return [Equals(sl, sr)]

        elif formula.nodeType == FNode.LE:
            sl = formula.left
            sr = formula.right
            return [Equals(sl, sr)]

        elif formula.nodeType == FNode.EQUALS:
            return [formula]

        elif formula.nodeType == FNode.LT:
            return []

        elif formula.nodeType == FNode.GT:
            return []

        else:
            assert False


    def extractStrategy(self, clausifier, gamma=None, psi=None):
        #assert (gamma and psi) or clausifier.problem.disjunctivity == DisjunctivityLevel.NONE

        psisolver = MathSAT5Solver(self.environment, "/tmp/log.smt2")

        st = self.environment.symbolTable
        yvars = clausifier.yvars
        xvars = clausifier.xvars

        print("Starting strategy extraction...")

        s = PiecewiseStrategy(yvars)

        print("Getting initial uncontrollable region...")
        if gamma is None:
            gamma = clausifier.getGamma()

        if psi is None:
            psi = clausifier.getPsi()

        print("Declaring y variables...")
        for y in yvars:
            psisolver.declareVariable(y)

        print("Declaring x variables...")
        for x in xvars:
            psisolver.declareVariable(x)

        # internals = self.getInternals(And([psi, gamma]))

        print("  Declaring atom variables...")
        eqvars = []
        ceqvars = []
        allequalities = set()
        mutex = {}
        mutexsum = 0
        for a in clausifier.atoms:
            assert a.nodeType == FNode.GE or a.nodeType == FNode.LE
            equality = Equals(a.left, a.right)

            if equality not in allequalities:
                allequalities.add(equality)

                cqv = st.createNewFreshSymbol(Types.REAL, "_C_EQ_%d")
                psisolver.declareVariable(cqv)

                eqv = Equals(cqv, Constant(1))

                psisolver.addAssertion(Or([Not(equality), eqv]))
                psisolver.addAssertion(Or([equality, Equals(cqv, Constant(0))]))

                eqvars.append(eqv)
                ceqvars.append(cqv)

                if a.left in mutex:
                    if mutex[a.left] is not None:
                        f1 = Equals(cqv, Constant(0))
                        f2 = Equals(mutex[a.left], Constant(0))
                        psisolver.addAssertion(Or([f1, f2]))
                        mutexsum += 1
                        mutex[a.left] = None
                else:
                    mutex[a.left] = cqv

        costVar = Constant(0)
        for v in ceqvars:
            costVar = Plus(v, costVar)


        print("Asserting region...")
        psisolver.addAssertion(psi)
        print("Asserting Gamma")
        psisolver.addAssertion(gamma)

        print("Asserting not internal region...")
        psisolver.addAssertion(Or(eqvars))

        currentCost = len(eqvars) - mutexsum

        assumptions = []
        for i in xrange(0, currentCost+1):
            assum = st.createNewFreshSymbol(Types.BOOL, "_ASS_%d")
            psisolver.declareVariable(assum)
            psisolver.addAssertion(Iff(assum, GE(costVar, Constant(i))))
            assumptions.append(assum)

        while currentCost > 0:
            print("\n  Checking if the remaining region is feasable...")

            if not psisolver.solve():
                print("  Region is infeasable, therefore I covered all the space!")
                print("  Done!")
                psisolver.destroy()
                return s

            print("  Checking if I can satisfy %s constraints..." % currentCost)

            if psisolver.solveAssuming(assumptions[currentCost]):
                print("  Region is feasible")
                print("    The model is:")
                xmodel = {}
                for x in xvars+yvars:
                    val = psisolver.getModelValue(x)
                    xmodel[x] = val

                print("  Getting the equality system")
                conj = []
                for v in allequalities:
                    val = psisolver.getModelValue(v)
                    if val == True:
                        conj.append(v)

                assert len(conj) > 0
                system = And(conj)

                (p, ker) = self.getLinearStrategyFromSystem(system,
                                                            clausifier,
                                                            xmodel)
                if ker >= len(yvars):
                    beta = self.getPsiWithStartegy(p, clausifier)
                    beta = beta.simplify(True)

                    if beta.nodeType == FNode.BCONSTANT and not beta.value:
                        assert False

                    s.addPiece(beta, p)
                    psisolver.addAssertion(Not(beta))

                psisolver.addAssertion(Not(system))

            else:
                currentCost -= 1


        return None
    pass
