from encoders.helpers import *
from extractors.strategy import *

from extractors.solvers.msat import *
from checkers import *

import common
import random


class LinearWeakStrategyExtractor(object):

    def __init__(self, quantelim, environment):
        self.quantelim = quantelim
        self.environment = environment

        self.solver = MathSAT5Solver(self.environment)
        self.checker = SyntacticChecker()
        return


    def extractStrategy(self, clausifier):
        st = self.environment.symbolTable

        yvars = clausifier.yvars
        xvars = clausifier.xvars

        s = LinearStrategy(yvars)

        print("Starting strategy extraction...")

        print("  Declaring Variables...")
        coeff = []
        for i in range(0, len(yvars) + 1):
            nv = st.createNewFreshSymbol(Types.REAL, "_C_%d")
            coeff.append(nv)
            self.solver.declareVariable(nv)

        done = []
        for x in xvars:
            if len(done) == 0:
                print("  Declaring '" + x.name + "' the zero starting point...")
                s.addZeroPolicy(x)
            else:
                print("  Building strategy for '" + x.name + "'...")
    
                self.solver.push()
                print("    Encoding problem as SMT formula...")
                formula = self.encode(clausifier, s, x, coeff, done)

                print("    Asserting SMT formula...")
                self.solver.addAssertion(formula)

                print("    Solving SMT formula...")
                if not self.solver.solve():
                    common.error("No linear strategy exists!")
                    exit(1)
                
                print("    Extracting model...")
                vals = []
                for c in coeff:
                    vals.append(self.solver.getModelValue(c))
                self.solver.pop()
                s.addPolicy(x, vals)
    
            done.append(x)
        print("Done.")
        return s


    def encode(self, clausifier, policy, pivot, coeff, completed):
        xvars = clausifier.xvars
        yvars = clausifier.yvars

        quantified = set(xvars)
        quantified.discard(pivot)
        
        print("      Preparing formula definitions...")
        formula = clausifier.getPsi().simplify()

        for x in completed:
            quantified.discard(x)
            formula = Definition(x.name, policy.getPolicyAsFormula(x), formula)

        formula = formula.simplify()
        
        if len(quantified) > 0:
            print("      Perform Quantifier Elimination...")
            formula = Exists(quantified, formula)
            elim = self.quantelim.removeQuantifiers(formula)
            elim = elim.simplify()
        else:
            elim = formula

        # if self.checker.checkTrue(elim):
        #     print("        The quantification is trivially true.")
        #     return TextFormula("true", [])

        if self.checker.checkFalse(elim):
            print("        The quantification is trivially false.")
            return TextFormula("false", [])

        ycount = len(yvars)

        print("      Building problem formula...")
        f = coeff[-1]
        for i in range(0, ycount):
            c = coeff[i]
            f = Plus(Times(c, yvars[i]), f)

        mass = Definition(pivot.name, f, elim)
        mass = mass.simplify()

        resList = self.allCases(clausifier, list(yvars), mass)
        return And(resList)


    def allCases(self, clausifier, choices, base, count=0, total=None):
        if len(choices) == 0:
            return [base]

        pivot = choices.pop()
        (l,u) = clausifier.getYBound(pivot)

        if total is None:
            total = len(choices)
        count = count+1
        
        sonList = self.allCases(clausifier, choices, base, count, total)

        perc = 100 / (2**count)

        print("        [%d/%d]\tGenerating cases for variable %s.\t\t(%.2f%% overall completion)" % (total-count+2, total+1, pivot.name, perc))
        res = []
        for s in sonList:
            res.append(Definition(pivot.name, Constant(l), s))
            res.append(Definition(pivot.name, Constant(u), s))

        return res
  
    pass
