#!/usr/bin/env python

import argparse
import os.path
import sys

sys.path.append(os.path.join(os.path.dirname(sys.argv[0]), "lib"))
sys.path.append(os.path.join(os.path.dirname(sys.argv[0]), "bin", "pylib"))
sys.setrecursionlimit(100000)

from structures import Problem, AtomicConstraint
from typing import Types
from formula import Environment
from extractors.solvers.msat import *


class PVYS(object):

    def __init__(self, environment):
        self.environment = environment
        self.solver = None
        return


    def solve(self, problem):
        st = self.environment.symbolTable

        Cs = set()
        Cc = set()
        Ce = set()

        for c in problem.constraints:
            if c.isSimple() or not c.hasUncontrollables():
                # all simple constraints + constraints on controllable variables
                Cs.add(c)
            elif c.contingent:
                # Disjunctive contingent constraints
                Cc.add(c)
            else:
                # Disjunctive free constraints with at least one uncontrollable
                Ce.add(c)


        self.solver = MathSAT5Solver(self.environment)
        for tp in problem.timePoints.values():
            v = st.createNewSymbol(tp.name, Types.REAL)
            self.solver.declareVariable(v)

        res = self.dtpu_sc(set(), set(), Cs, Cc, Ce)

        self.solver.destroy()
        self.solver = None

        return res

    def select_variable(self, s):
        assert len(s) > 0
        for v in s:
            return v


    def disjuncts(self, c):
        return c.atoms


    def sc_transform(self, prob, c):
        print "-" * 10
        print "  Input: prob=", prob, " , c=", c, "(contingent=", c.contingent , ")"
        def get_contingency(t):
            for a in prob:
                if a.contingent and a.dest == t:
                    return (a.source, a.lowerBound, a.upperBound)
            return (None, None, None)

        def plus(a, b):
            assert b != common.INFINITY and b != common.MINUS_INFINITY
            if a == common.INFINITY or a == common.MINUS_INFINITY:
                return a

            return a + b


        toprocess = []
        if c.contingent:
            for a in prob:
                if a != c and (a.source == c.dest or a.dest == c.dest):
                    toprocess.append(a)
        else:
            toprocess.append(c)

        result = set()
        print "  ToProcess=", toprocess
        for p in toprocess:
            if p.source.controllable and p.dest.controllable:
                print "A"
                result.add(c)
            elif p.dest.controllable:
                # Free of type 2
                (s, l, u) = get_contingency(p.source)
                if s is not None:
                    res = AtomicConstraint(s, p.dest,
                                           plus(p.lowerBound, u),
                                           plus(p.upperBound, l),
                                           False)
                    result.add(res)

            elif p.source.controllable:
                # Free of type 3
                (s, l, u) = get_contingency(p.dest)
                if s is not None:
                    res = AtomicConstraint(p.source, s,
                                           plus(p.lowerBound, -l),
                                           plus(p.upperBound, -u),
                                           False)
                    result.add(res)

            else:
                # Free of type 4
                (s1, l1, u1) = get_contingency(p.dest)
                (s2, l2, u2) = get_contingency(p.source)
                print l1, u1, l2, u2
                if s1 is not None and s2 is not None:
                    res = AtomicConstraint(s1, s2,
                                           plus(p.lowerBound, u2 - l1),
                                           plus(p.upperBound, l2 - u1),
                                           False)
                    result.add(res)

        print "  Output: ", result
        print "-" * 10
        return result


    def minimal_network(self, prob):
        def lt(a, b):
            if a == common.INFINITY:
                return False
            if b == common.INFINITY:
                return True
            return a < b

        def plus(a, b):
            if a == common.INFINITY or b == common.INFINITY:
                return common.INFINITY
            return a + b


        tps = list(set([a.source for a in prob]) | set([a.dest for a in prob]))
        M = [[common.INFINITY for _ in tps] for _ in tps]

        ids = {}
        c = 0
        for t in tps:
            ids[t] = c
            M[c][c] = 0
            c += 1

        for a in prob:
            s = a.source
            d = a.dest
            l = a.lowerBound
            u = a.upperBound

            if lt(u, M[ids[s]][ids[d]]):
                M[ids[s]][ids[d]] = u
            if l != common.MINUS_INFINITY and lt(-l, M[ids[d]][ids[s]]):
                M[ids[d]][ids[s]] = -1 * l

        for k in tps:
            for i in tps:
                for j in tps:
                    if lt(plus(M[ids[i]][ids[k]], M[ids[k]][ids[j]]), M[ids[i]][ids[j]]):
                         M[ids[i]][ids[j]] = plus(M[ids[i]][ids[k]], M[ids[k]][ids[j]])

        R = set()

        c = 0
        for i in tps:
            for j in tps[(c+1):]:
                assert i != j
                u = M[ids[i]][ids[j]]
                l = M[ids[j]][ids[i]]

                if u != common.INFINITY or l != common.INFINITY:
                    if l == common.INFINITY:
                        l = common.MINUS_INFINITY
                    else:
                        l = l * -1
                    a = AtomicConstraint(i, j, l, u)
                    R.add(a)

            c += 1

        return R


    def is_consistent(self, prob):
        assert self.solver is not None
        st = self.environment.symbolTable

        if len(prob) == 0:
            return True

        self.solver.push()

        for a in prob:
            vs = st.getSymbolByName(a.source.name)
            vd = st.getSymbolByName(a.dest.name)
            l = a.lowerBound
            u = a.upperBound

            if l != common.MINUS_INFINITY:
                self.solver.addAssertion(GE(Minus(vd,vs), Constant(l)))
            if u != common.INFINITY:
                self.solver.addAssertion(LE(Minus(vd,vs), Constant(u)))

        res = self.solver.solve()

        self.solver.pop()

        return res


    def dtpu_sc(self, A, Ac, Cs, Cc, Ce):
        S = set()
        if len(Cs) == 0:
            G = self.minimal_network(A)
            print "Minimal Network:", G
            S = self.all_paths_sc(A, Ac, Cc, Ce, G)
        else:
            Ci = self.select_variable(Cs)
            print "Picked %s from %s" % (str(Ci), str(Cs))

            Cs1 = Cs - set([Ci])

            for cij in self.disjuncts(Ci):
                Ac1 = Ac | set([cij])
                A1 = A | self.sc_transform(Ac1, cij)
                print "A1:", A1
                if self.is_consistent(A1):
                    S = self.dtpu_sc(A1, Ac1, Cs1, Cc, Ce)
                    if len(S) > 0:
                        return S
        return S


    def all_paths_sc(self, A, Ac, Cc, Ce, G):
        if len(Cc) == 0:
            print "G:", G
            G = G & self.satisfy_ce(A, Ac, Ce)
            print "G':", G
        else:
            Ci = self.select_variable(Cc)
            Cc1 = Cc - set([Ci])
            for cij in self.disjuncts(Ci):
                Ac1 = Ac | set([cij])
                A1 = A | self.sc_transform(Ac1, cij)
                if self.is_consistent(A1):
                    G = self.all_paths_sc(A1, Ac1, Cc1, Ce, G)
                    if len(G) > 0:
                        return set()
                else:
                    return set()
        return G


    def satisfy_ce(self, A, Ac, Ce):
        H = set()
        if len(Ce) == 0:
            H = self.minimal_network(A)
            print "H=", H
        else:
            Ci = self.select_variable(Ce)
            Ce1 = Ce - set([Ci])
            for cij in self.disjuncts(Ci):
                print "A:", A
                A1 = A | self.sc_transform(Ac, cij)
                print "A':", A1
                if self.is_consistent(A1):
                    print "consistent!"
                    H = H | self.satisfy_ce(A1, Ac, Ce1)
        return H

#End of PVYS




# Main script
if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='PVYS Strong Controllability Solver')

    parser.add_argument('inputFile', metavar='inputFile', type=str,
                        help='The probmem to solve')

    args = parser.parse_args()




    env = Environment()

    tn = Problem()
    tn.fromFile(args.inputFile)


    s = PVYS(env)
    print s.solve(tn)
